/* crunch - header file with runtime support */

#ifndef CRUNCH_H
#define CRUNCH_H

#ifdef CRUNCH_FREESTANDING
#include <inttypes.h>
#include <stdarg.h>
#include <complex.h>
#include "crunch-environment.h"
#else
#define _STD_FORMAT_MACROS
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <complex.h>
#include <assert.h>
#include <math.h>
#include <time.h>
#include <sys/time.h>
#include <unistd.h>
#include <sys/stat.h>
#include <errno.h>
#include <inttypes.h>
#include <stdarg.h>
#endif

#ifdef CRUNCH_EMBEDDED
#define CRUNCH_STATIC	static
#define CRUNCH_EXTERN	static
#else
#define CRUNCH_STATIC
#define CRUNCH_EXTERN	extern
#endif

#if defined(__LP64__) || defined(_LP64) || defined(__MINGW64__) || defined(_WIN64)
# define CRUNCH_64	1
# define crunch_word		int64_t
# define crunch_halfword	int32_t
# define crunch_intvector	crunch_s64vector
# define crunch_write_intvector	crunch_write_s64vector
# define crunch_intvector_block	crunch_s64vector_block
# define crunch_intvector_append	crunch_s64vector_append
# define crunch_make_intvector(n, x)	crunch_make_s64vector(2, (n), (x))
# define crunch_cons_intvector	crunch_cons_s64vector
# define CRUNCH_PRSIZE	PRIi64
#else
# define crunch_word		int32_t
# define crunch_halfword	int16_t
# define crunch_intvector	crunch_s32vector
# define crunch_write_intvector	crunch_write_s32vector
# define crunch_intvector_block	crunch_s32vector_block
# define crunch_intvector_append	crunch_s32vector_append
# define crunch_make_intvector(n, x)	crunch_make_s64vector(2, (n), (x))
# define crunch_cons_intvector	crunch_cons_s32vector
# define CRUNCH_PRSIZE	PRIi32
#endif

#ifdef CRUNCH_DEBUG
# define crunch_log(fmt, ...)		fprintf(stderr, fmt, __VA_ARGS__)
#else
# define crunch_log(fmt, ...)
#endif

#define CRUNCH_BUFFER_SIZE	32000

/* types */

#define CRUNCH_NESTEDRC	0x0001
#define CRUNCH_PORT		0x0002
#define CRUNCH_TEXTPORT	0x0004
#define CRUNCH_EOF			0x0008
#define CRUNCH_INPUTPORT	0x0010
#define CRUNCH_OUTPUTPORT	0x0020
#define CRUNCH_REFPORT	0x0040
#define CRUNCH_OPEN		0x0080
#define CRUNCH_PINNED		0x0100
#define CRUNCH_BLOCK		0x0200

#ifdef CRUNCH_DEBUG
# define crunch_sequence(t) \
struct { \
	crunch_halfword rc; \
	crunch_halfword flags; \
	crunch_word len; \
	crunch_word size; \
	t *data; \
	void *next, *prev; \
}
#else
# define crunch_sequence(t) \
struct { \
	crunch_halfword rc; \
	crunch_halfword flags; \
	crunch_word len; \
	crunch_word size; \
	t *data; \
}
#endif

#ifdef CRUNCH_NO_UTF
# define crunch_charvector	crunch_u8vector
#else
# define crunch_charvector	crunch_u32vector
#endif

typedef crunch_sequence(char) crunch_string_block;
typedef crunch_sequence(void) crunch_vector_block;
typedef crunch_sequence(uint8_t) crunch_u8vector_block;
typedef crunch_sequence(int8_t) crunch_s8vector_block;
typedef crunch_sequence(uint16_t) crunch_u16vector_block;
typedef crunch_sequence(int16_t) crunch_s16vector_block;
typedef crunch_sequence(uint32_t) crunch_u32vector_block;
typedef crunch_sequence(int32_t) crunch_s32vector_block;
typedef crunch_sequence(uint64_t) crunch_u64vector_block;
typedef crunch_sequence(int64_t) crunch_s64vector_block;
typedef crunch_sequence(float) crunch_f32vector_block;
typedef crunch_sequence(double) crunch_f64vector_block;
typedef crunch_sequence(float complex) crunch_c64vector_block;
typedef crunch_sequence(double complex) crunch_c128vector_block;

typedef struct crunch_port_info {
	FILE *file;
	crunch_vector_block *ref;
	long peeked;
} crunch_port_info;

typedef crunch_sequence(crunch_port_info) crunch_port_block;

typedef crunch_string_block *crunch_string;
typedef crunch_vector_block *crunch_vector;
typedef crunch_u8vector_block *crunch_u8vector;
typedef crunch_s8vector_block *crunch_s8vector;
typedef crunch_u16vector_block *crunch_u16vector;
typedef crunch_s16vector_block *crunch_s16vector;
typedef crunch_u32vector_block *crunch_u32vector;
typedef crunch_s32vector_block *crunch_s32vector;
typedef crunch_u64vector_block *crunch_u64vector;
typedef crunch_s64vector_block *crunch_s64vector;
typedef crunch_f32vector_block *crunch_f32vector;
typedef crunch_f64vector_block *crunch_f64vector;
typedef crunch_c64vector_block *crunch_c64vector;
typedef crunch_c128vector_block *crunch_c128vector;
typedef crunch_port_block *crunch_port;

typedef void *crunch_function;

/* environment setup */

static struct crunch_process_context {
	int argc;
	char **argv;
	char **envp;
#ifdef CRUNCH_DEBUG
	crunch_vector_block *tracked;
#endif
} crunch_process_context;

static crunch_port_info crunch_stdout_info = { NULL, NULL, -1 };
static crunch_port_info crunch_stdin_info = { NULL, NULL, -1 };
static crunch_port_info crunch_stderr_info = { NULL, NULL, -1 };
static crunch_port_block crunch_stdout_block = { -1, CRUNCH_PORT | CRUNCH_TEXTPORT | CRUNCH_OUTPUTPORT, 0, 0, &crunch_stdout_info };
static crunch_port_block crunch_stderr_block = { -1, CRUNCH_PORT | CRUNCH_TEXTPORT | CRUNCH_OUTPUTPORT, 0, 0, &crunch_stdin_info };
static crunch_port_block crunch_stdin_block = { -1, CRUNCH_PORT | CRUNCH_TEXTPORT | CRUNCH_INPUTPORT, 0, 0, &crunch_stderr_info };
static crunch_port crunch_stdout = &crunch_stdout_block;
static crunch_port crunch_stderr = &crunch_stderr_block;
static crunch_port crunch_stdin = &crunch_stdin_block;

static void crunch_ensure_stdio(void) {
	if(crunch_stdout->data->file == NULL) {
		crunch_stdout->data->file = stdout;
		crunch_stdin->data->file = stdin;
		crunch_stderr->data->file = stderr;
	}
}

static void crunch_main(int argc, char **argv, char **envp) {
	crunch_process_context.argc = argc;
	crunch_process_context.argv = argv;
	crunch_process_context.envp = envp;
#ifdef CRUNCH_FREESTANDING
	crunch_init_host();		// provided by user
#endif
#ifdef CRUNCH_DEBUG
	crunch_process_context.tracked = NULL;
#endif
}

#define crunch_optional(n)	(argc <= (n))

#define crunch_identity(x)	(x)

static inline void crunch_void(int n, ...) {}

static void crunch_abort(char *msg) {
	fprintf(stderr, "Runtime error: %s\n", msg);
	exit(1);
}

static inline void crunch_error(int n, crunch_string msg, ...) {
	crunch_abort(msg->data);
}

static crunch_vector_block *crunch_alloc(size_t size) {
	crunch_vector_block *p = malloc(sizeof(crunch_vector_block));
	assert(p != NULL);
	p->size = size;
	p->flags = 0;
	p->rc = 0;
#ifdef CRUNCH_DEBUG
	if(crunch_process_context.tracked != NULL)
		crunch_process_context.tracked->prev = p;
	p->next = crunch_process_context.tracked;
	p->prev = NULL;
	crunch_process_context.tracked = p;
#endif
	p->data = malloc(size);
	assert(p->data);
	memset(p->data, 0, size);
	return p;
}

#define crunch_pin(v)		(((crunch_vector_block *)(v))->flags |= CRUNCH_PINNED)

static void crunch_pin_rec(crunch_vector_block *b) {
	crunch_pin(b);
	if((b->flags & CRUNCH_NESTEDRC) != 0) {
		crunch_vector_block **b2 = (crunch_vector_block **)b->data;
		for(int i = 0; i < b->len; ++i) crunch_pin_rec(*(b2++));
	}
}

static void crunch_cleanup(void) {
#ifdef CRUNCH_DEBUG
	// mark nested
	for(crunch_vector_block *p = crunch_process_context.tracked; p != NULL; p = p->next) {
		if((p->flags & CRUNCH_PINNED) != 0) crunch_pin_rec(p);
	}
	for(crunch_vector_block *p = crunch_process_context.tracked; p != NULL; p = p->next) {
		if((p->flags & CRUNCH_PINNED) == 0)
			fprintf(stderr, "memory leak for block %p (size %" CRUNCH_PRSIZE ") with reference count %d (flags: 0x%04x)\n",
				p, p->size, p->rc, p->flags);
	}
#endif
}


/* refcounting */

#ifdef CRUNCH_ATOMIC_REFCOUNTS
#define CRUNCH_INCRC(rc)	__sync_add_and_fetch(&(rc), 1)
#define CRUNCH_DECRC(rc)	__sync_sub_and_fetch(&(rc), 1)
#else
#define CRUNCH_INCRC(rc)	(++(rc))
#define CRUNCH_DECRC(rc)	(--(rc))
#endif

static inline void crunch_addref_internal(crunch_vector_block *v)
{
	if(v->rc > -1) CRUNCH_INCRC(v->rc);
}

static inline void crunch_release(crunch_vector_block *v) {
	if((v->flags & CRUNCH_BLOCK) == 0) free(v->data);
	free(v);
}

static inline void crunch_unref_internal(crunch_vector_block *v)
{
	if(v == NULL) return;
	if(v->rc > -1) {
		if(CRUNCH_DECRC(v->rc) == 0) {
			if((v->flags & CRUNCH_NESTEDRC) != 0) {
				for(int i = 0; i < v->len; ++i)
					crunch_unref_internal(((crunch_vector_block **)v->data)[ i ]);
			}
#ifdef CRUNCH_DEBUG
			if(v->prev == NULL) crunch_process_context.tracked = v->next;
			else ((crunch_vector_block *)v->prev)->next = v->next;
			if(v->next != NULL) ((crunch_vector_block *)v->next)->prev = v->prev;
#endif
			if((v->flags & CRUNCH_PORT) != 0) {
				crunch_port_info *info = (crunch_port_info *)v->data;
				if((v->flags & CRUNCH_OPEN) != 0) fclose(info->file);
				if((v->flags & CRUNCH_REFPORT) != 0)
					crunch_unref_internal(info->ref);
			}
			crunch_release(v);
		}
	}
}

#define crunch_addref(v)	(crunch_addref_internal((crunch_vector_block *)(v)))
#define crunch_unref(v)		(crunch_unref_internal((crunch_vector_block *)(v)))

#define crunch_unref_result(v)	(CRUNCH_DECRC((v)->rc), (v))

#define crunch_move(t, from, to)	{t _tmp = (t)from; crunch_addref(_tmp); crunch_unref(to); to = _tmp;}

#ifdef CRUNCH_DEBUG
# define crunch_move_and_pin(t, from, to)	{t _tmp = (t)from; crunch_addref(_tmp); crunch_unref(to); crunch_pin(_tmp); to = _tmp;}
#else
# define crunch_move_and_pin	crunch_mode
#endif

/* predefined literals and initializers */

static crunch_vector_block crunch_null_string = {-1, 0, 1, 0, ""};
static crunch_vector_block crunch_null_vector = {-1, 0, 0, 0, NULL};
static crunch_vector_block crunch_eof_vector = {-1, CRUNCH_EOF, 0, 0, NULL };
static crunch_vector_block crunch_null_port = {-1, CRUNCH_PORT, 0, 0, NULL };

#define crunch_abstract_procedure 	((void *)0xfacade)

#define crunch_eof_scalar	(-1)

#define crunch_eofp(v)	(((v)->flags & CRUNCH_EOF) != 0)

#ifndef CRUNCH_NO_UTF
CRUNCH_EXTERN char *crunch_utf_decode(char *buf, uint32_t *c);
CRUNCH_EXTERN char *crunch_utf_encode(uint32_t c, char *p);
CRUNCH_EXTERN int crunch_utf_isalpha(uint32_t c);
CRUNCH_EXTERN int crunch_utf_isspace(uint32_t c);
CRUNCH_EXTERN int crunch_utf_islower(uint32_t c);
CRUNCH_EXTERN int crunch_utf_isupper(uint32_t c);
CRUNCH_EXTERN int crunch_utf_isdigit(uint32_t c);
CRUNCH_EXTERN int crunch_utf_count(char *p, int len);
CRUNCH_EXTERN uint32_t crunch_utf_char_downcase(uint32_t c);
CRUNCH_EXTERN uint32_t crunch_utf_char_upcase(uint32_t c);
CRUNCH_EXTERN uint32_t crunch_utf_char_foldcase(uint32_t c);
CRUNCH_EXTERN int crunch_utf_compare(crunch_string s1, crunch_string s2, crunch_word start1, crunch_word start2, crunch_word len);
CRUNCH_EXTERN int crunch_utf_compare_ci(crunch_string s1, crunch_string s2, crunch_word start1, crunch_word start2, crunch_word len);
CRUNCH_EXTERN char *crunch_utf_index(crunch_string s, crunch_word p);
#endif	// !CRUNCH_NO_UTF

#define crunch_data(v)		((v)->data)

/* (scheme base) */

#ifdef CRUNCH_NO_UTF
static inline int crunch_utf_isalpha(uint32_t c) {
    return (c >= 'a' && c <= 'z') || (c >= 'A' && c <= 'Z');
}

static inline int crunch_utf_isspace(uint32_t c) {
    return c == ' ' || c == 9 || c == '\t';  //XXX more?
}

static inline int crunch_utf_isupper(uint32_t c) {
    return (c >= 'A' && c <= 'Z');
}

static inline int crunch_utf_islower(uint32_t c) {
    return (c >= 'a' && c <= 'z');
}

static inline int crunch_utf_isdigit(uint32_t c) {
    return (c >= '0' && c <= '9');
}

static inline uint32_t crunch_utf_char_downcase(uint32_t c) {
    return crunch_utf_isupper(c) ? c + 32 : c;
}

static inline uint32_t crunch_utf_char_upcase(uint32_t c) {
    return crunch_utf_islower(c) ? c - 32 : c;
}

#define crunch_utf_char_foldcase	crunch_utf_char_downcase
#endif	// CRUNCH_NO_UTF

#define crunch_char_to_integer(x)	((crunch_word)(x))
#define crunch_integer_to_char(x)	((uint32_t)(x))

#define crunch_equalp(x, y)	\
	(crunch_equalp_internal((crunch_vector_block *)(x), (crunch_vector_block *)(y)))

static inline int crunch_equalp_internal(crunch_vector_block *x, crunch_vector_block *y) {
 	if(x == y) return 1;
	if(x->size != y->size) return 0;
	if((x->flags & CRUNCH_NESTEDRC) != 0) {
		if((y->flags & CRUNCH_NESTEDRC) == 0) return 0;
		for(int i = 0; i < x->len; ++i) {
			if(!crunch_equalp_internal(((crunch_vector_block **)x->data)[ i ],
				((crunch_vector_block **)y->data)[ i ]))
				return 0;
		}
		return 1;
	} else return memcmp(x->data, y->data, x->size) == 0;
}

static long crunch_integer_expt(long x, long n) {
	long r = 1;
	while(n > 0) {
		if((n & 1) == 1) r *= x;
		x *= x;
		n >>= 1;
	}
	return r;
}

#define crunch_inverse(x)  (1 / (x))

static inline void crunch_newline(int argc, ...) {
	va_list va;
	va_start(va, argc);
	crunch_port port = crunch_stdout;
	crunch_ensure_stdio();
	if(argc > 0) port = va_arg(va, crunch_port);
	fputc('\n', port->data->file);
}

static inline long crunch_string_length(crunch_string s) {
#ifdef CRUNCH_NO_UTF
	return(s->size);
#else
	return(s->len);
#endif
}

static long crunch_remainder(long x, long y) {
	return x - ((x / y) * y);
}

static long crunch_modulo(long x, long y) {
	x = x % y;
	if ((y < 0 && x > 0) || (y > 0 && x < 0)) x += y;
	return x;
}

#define crunch_square(x)	((x) * (x))

#define crunch_max(x, y)	((x) > (y) ? (x) : (y))
#define crunch_min(x, y)	((x) < (y) ? (x) : (y))

#define crunch_vector_length(v)	(((crunch_vector)v)->len)

static crunch_string crunch_cons_string(int n, ...) {
	va_list va;
	va_start(va, n);
#ifdef CRUNCH_NO_UTF
	crunch_string s = (crunch_string)crunch_alloc(n + 1);
	for(int i = 0; i < n; ++i) s->data[ i ] = va_arg(va, uint32_t);
#else
	crunch_string s = (crunch_string)crunch_alloc(n * 4 + 1);
	char *p = s ->data;
	for(int i = 0; i < n; ++i)
		p = crunch_utf_encode(va_arg(va, uint32_t), p);
	int len = p - s->data;
	s->data = realloc(s->data, len + 1);
	s->data[ len ] = '\0';
	s->size = len + 1;
#endif
	s->len = n;
	return s;
}

#define crunch_make_cons_numvector(t, et, at) \
static crunch_ ## t ## vector crunch_cons_ ## t ## vector(int n, ...) { \
	va_list va; \
	va_start(va, n); \
	crunch_ ## t ## vector v = (crunch_ ## t ## vector)crunch_alloc(sizeof(et) * n); \
	for(int i = 0; i < n; ++i) v->data[ i ] = va_arg(va, at); \
	v->len = n; \
	return v; \
}

crunch_make_cons_numvector(u8, uint8_t, long)
crunch_make_cons_numvector(u32, uint32_t, long)
crunch_make_cons_numvector(s32, int32_t, long)
crunch_make_cons_numvector(s64, int64_t, long)
crunch_make_cons_numvector(f64, double, double)
crunch_make_cons_numvector(c128, double complex, double complex)

static crunch_vector crunch_cons_vector(int n, ...) {
	va_list va;
	va_start(va, n);
	crunch_vector v = (crunch_vector)crunch_alloc(sizeof(crunch_vector_block *) * n);
	for(int i = 0; i < n; ++i)
		((crunch_vector_block **)v->data)[ i ] = va_arg(va, crunch_vector_block *);
	v->len = n;
	return v;
}

static inline int crunch_string_equal(crunch_string s1, crunch_string s2) {
	return s1->size == s2->size && memcmp(s1->data, s2->data, s1->size) == 0;
}

static inline int crunch_string_compare(crunch_string s1, crunch_string s2) {
	int sz = s1->size < s2->size ? s1->size : s2->size;
#ifdef CRUNCH_NO_UTF
	return memcmp(s1->data, s2->data, sz);
#else
	return crunch_utf_compare(s1, s2, 0, 0, sz);
#endif
}

static inline int crunch_string_compare_ci(crunch_string s1, crunch_string s2) {
	int sz = s1->size < s2->size ? s1->size : s2->size;
#ifdef CRUNCH_NO_UTF
	return strncasecmp(s1->data, s2->data, sz);
#else
	return crunch_utf_compare_ci(s1, s2, 0, 0, sz);
#endif
}

static inline int crunch_string_less(crunch_string s1, crunch_string s2) {
	int r = crunch_string_compare(s1, s2);
	if(r == 0) return s1->size < s2->size;
	return r < 0;
}

static inline int crunch_string_greater(crunch_string s1, crunch_string s2) {
	int r = crunch_string_compare(s1, s2);
	if(r == 0) return s1->size > s2->size;
	return r > 0;
}

static inline int crunch_string_less_or_equal(crunch_string s1, crunch_string s2) {
	int r = crunch_string_compare(s1, s2);
	if(r == 0) return s1->size <= s2->size;
	return r <= 0;
}

static inline int crunch_string_greater_or_equal(crunch_string s1, crunch_string s2) {
	int r = crunch_string_compare(s1, s2);
	if(r == 0) return s1->size >= s2->size;
	return r >= 0;
}

static inline void crunch_bounds_check_internal(int len, long i) {
	assert(i >= 0 && i < len);
}

#define crunch_bounds_check(v, i)	crunch_bounds_check_internal((v)->len, (i))
#define crunch_numvector_ref(v, i)	(crunch_bounds_check((v), (i)), ((v)->data[ i ]))
#define crunch_vector_ref(v, i)	(crunch_bounds_check((v), (i)), ((crunch_vector_block **)(v)->data)[ i ])

static crunch_vector crunch_vector_copy_internal(int n, crunch_vector_block *v, ...) {
   	va_list va;
	int from = 0, to, esz = v->size / v->len;
	va_start(va, v);
	if(n > 1) from = va_arg(va, long);
	if(n > 2) to = va_arg(va, long);
	else to = v->len;
	assert(from >= 0 && to >= 0 && from <= v->len && to <= v->len); \
	crunch_vector v2 = (crunch_vector)crunch_alloc(esz * (to - from));
	v2->flags = v->flags;
	v2->len = to - from;
	memcpy(v2->data, v->data, v2->len * esz);
	if((v2->flags & CRUNCH_NESTEDRC) != 0) {
		for(int i = 0; i < v2->len; ++i)
			crunch_addref(((crunch_vector_block **)v2->data)[ i  ]);
	}
	return v2;
}

#define crunch_vector_copy1(n, v)	crunch_vector_copy_internal((n), (crunch_vector_block *)(v))
#define crunch_vector_copy(n, v, ...)	\
	crunch_vector_copy_internal((n), (crunch_vector_block *)(v) __VA_ARGS__)

#define crunch_numvector_set(v, i, x)		(crunch_bounds_check((v), (i)), (v)->data[ i ] = (x))

#define crunch_vector_set(v, i, x)	\
	crunch_move(crunch_vector_block *, (x), ((crunch_vector_block **)(v)->data)[ i ])

static crunch_u8vector crunch_u8vector_copy1(crunch_u8vector v) {
	crunch_u8vector v2 = (crunch_u8vector)crunch_alloc(v->size);
	v2->len = v->len;
	memcpy(v2->data, v->data, v->size);
	return v2;
}

#define crunch_make_numvector_copy(t) \
static crunch_ ## t ## vector crunch_ ## t ## vector_copy(int n, crunch_ ## t ## vector v, long from, ...) { \
	va_list va; int to; \
	va_start(va, from); \
	if(n > 2) to = va_arg(va, long); \
	else to = v->len; \
	int esz = v->size / v->len; \
	assert(from >= 0 && to >= 0 && from <= v->len && to <= v->len); \
	crunch_ ## t ## vector v2 = (crunch_ ## t ## vector)crunch_alloc((to - from) * esz); \
	v2->len = to - from; \
	memcpy(v2->data, v->data + from, v2->size = v2->len * esz); \
	return v2; \
}

crunch_make_numvector_copy(u8)

#define crunch_make_subnumvector(t) \
static crunch_ ## t ## vector crunch_sub ## t ## vector(crunch_ ## t ## vector v, long from, long to) { \
	int esz = v->size / v->len; \
	assert(from >= 0 && to >= 0 && from <= v->len && to <= v->len); \
	crunch_ ## t ## vector v2 = (crunch_ ## t ## vector)crunch_alloc((to - from) * esz); \
	v2->len = to - from; \
	memcpy(v2->data, v->data + from, v2->size = v2->len * esz); \
	return v2; \
}

crunch_make_subnumvector(u8)
crunch_make_subnumvector(s8)
crunch_make_subnumvector(u16)
crunch_make_subnumvector(s16)
crunch_make_subnumvector(u32)
crunch_make_subnumvector(s32)
crunch_make_subnumvector(u64)
crunch_make_subnumvector(s64)
crunch_make_subnumvector(f32)
crunch_make_subnumvector(f64)
crunch_make_subnumvector(c64)
crunch_make_subnumvector(c128)

#define crunch_vector_copy_into(n, to, at, from, ...) \
	crunch_vector_copy_into_internal(n, (crunch_vector_block *)to, at, \
		(crunch_vector_block *)from, __VA_ARGS__)

static void crunch_vector_copy_into_internal(int n, crunch_vector_block *to, long at,
	crunch_vector_block *from, ...) {
	va_list va;
	va_start(va, from);
	int start = 0, end = from->len;
	if(n > 4) start = va_arg(va, long);
	if(n > 5) end = va_arg(va, long);
	int esz = from->size / from->len;
	int len = end - start;
	assert(at >= 0 && at < to->len);
	assert(to >= 0 && at + len < to->len);
	assert(start >= 0 && start < from->len && start < len);
	assert(end >= 0 && end <= from->len);
	if((to->flags & CRUNCH_NESTEDRC) != 0) {
		for(int i = start; i < end; ++i) {
			crunch_move(crunch_vector_block *,
				((crunch_vector_block **)from->data)[ i ],
				((crunch_vector_block **)to->data)[ at ]);
			++at;
		}
	} else memcpy(to->data + at * esz, from->data + esz * start, len * esz);
}

#define crunch_vector_fill(n, v, x, ...) \
	crunch_numvector_fill_internal(n, (crunch_vector_block *)(v), \
		(crunch_vector_block *)(x), __VA_ARGS__)

static void crunch_vector_fill_internal(int n, crunch_vector_block *v,
	crunch_vector_block *x, ...) {
	va_list va;
	va_start(va, x);
	int start = 0, end = v->len;
	if(n > 2) start = va_arg(va, long);
	if(n > 3) end = va_arg(va, long);
	assert(start >= 0 && start < v->len && start < end);
	assert(end >= 0 && end <= v->len);
	for(int i = start; i < end; ++i)
		crunch_move(crunch_vector_block *, x, ((crunch_vector_block **)v->data)[ i ]);
}

#define crunch_make_numvector_fill(vt, xt) \
static void crunch_ ## vt ## vector_fill(int n, crunch_ ## vt ## vector v, xt x, ...) { \
	va_list va; \
	va_start(va, x); \
	int start = 0, end = v->len; \
	if(n > 2) start = va_arg(va, long); \
	if(n > 3) end = va_arg(va, long); \
	assert(start >= 0 && start < v->len && start < end); \
	assert(end >= 0 && end <= v->len); \
	for(int i = start; i < end; ++i) v->data[ i ] = x; \
}

crunch_make_numvector_fill(u8, long)
crunch_make_numvector_fill(int, long)
crunch_make_numvector_fill(char, uint32_t)
crunch_make_numvector_fill(f64, double)
crunch_make_numvector_fill(c128, double complex)

#define crunch_make_vector(n, x) \
	crunch_make_vector_internal((n), (crunch_vector_block *)(x))

#define crunch_make_booleanvector(n, x)	crunch_make_u8vector(2, (n), (x))
#define crunch_make_floatvector(n, x)	crunch_make_f64vector(2, (n), (x))
#define crunch_make_complexvector(n, x)	crunch_make_c128vector(2, (n), (x))
#define crunch_make_charvector(n, x)	crunch_make_u32vector(2, (n), (x))

static crunch_vector_block *crunch_make_vector_internal(long n, crunch_vector_block *x) {
	assert(n >= 0);
	crunch_vector_block *v = crunch_alloc(n * sizeof(crunch_vector_block *));
	v->len = n;
	v->flags = CRUNCH_NESTEDRC;
	for(int i = 0; i< n; ++i)
		crunch_move(crunch_vector_block *, x, ((crunch_vector_block **)v->data)[ i ]);
	return v;
}

#define crunch_make_make_numvector(vt, xt, esz) \
static crunch_ ## vt ## vector crunch_make_ ## vt ## vector(int n, long s, ...) { \
	va_list va; \
	va_start(va, s); \
	xt x = 0; \
	assert(s >=0 ); \
	if(n > 1) x = va_arg(va, xt); \
	crunch_ ## vt ## vector v = (crunch_ ## vt ## vector)crunch_alloc(esz * s); \
	for(int i = 0; i < s; ++i) v->data[ i ] = x; \
	v->len = s; \
	return v; \
}

crunch_make_make_numvector(u8, long, sizeof(uint8_t))
crunch_make_make_numvector(s8, long, sizeof(int8_t))
crunch_make_make_numvector(u16, long, sizeof(uint16_t))
crunch_make_make_numvector(s16, long, sizeof(int16_t))
crunch_make_make_numvector(u32, long, sizeof(uint32_t))
crunch_make_make_numvector(s32, long, sizeof(int32_t))
crunch_make_make_numvector(u64, long, sizeof(uint64_t))
crunch_make_make_numvector(s64, long, sizeof(int64_t))
crunch_make_make_numvector(f32, double, sizeof(float))
crunch_make_make_numvector(f64, double, sizeof(double))
crunch_make_make_numvector(c64, float complex, sizeof(float complex))
crunch_make_make_numvector(c128, double complex, sizeof(double complex))

#define crunch_make_numvector_append(t) \
static crunch_ ## t ## vector crunch_ ## t ## vector_append(int n, crunch_ ## t ## vector v1, ...) { \
	crunch_ ## t ## vector v2, av[ 10 ]; \
	va_list va; \
	va_start(va, v1); \
	int i, total = v1->len, esz = v1->size / v1->len; \
	av[ 0 ] = v1; \
	for(i = 1; i < n; ++i) { \
		av[ i ] = va_arg(va, crunch_ ## t ## vector); \
		total += av[ i ]->len; \
	} \
	v2 = (crunch_ ## t ## vector)crunch_alloc(esz * total); \
	v2->len = total; \
	char *p = (char *)v2->data; \
	for(i = 0; i < n; ++i) { \
		memcpy(p, av[ i ]->data, av[ i ]->size); \
		p += av[ i ]->size; \
	} \
	return v2; \
}

crunch_make_numvector_append(u8)
crunch_make_numvector_append(s32)
crunch_make_numvector_append(s64)
crunch_make_numvector_append(f64)
crunch_make_numvector_append(c128)

static void crunch_u8vector_copy_into(int n, crunch_u8vector to,
	long at, crunch_u8vector from, ...) {
	va_list va;
	int start = 0, end;
	va_start(va, from);
	if(n > 3) start = va_arg(va, long);
	if(n > 4) end = va_arg(va, long);
	else end = from->len;
	memmove(to->data + at, from->data + start, end - start);
}

static crunch_vector crunch_vector_append(int n, crunch_vector v1, ...) {
	va_list va;
	va_start(va, v1);
	crunch_vector v2, av[ 10 ];
	int len = v1->len, i;
	av[ 0 ] = v1;
	for(i = 1; i < n; ++i) {
		av[ i ] = va_arg(va, crunch_vector);
		len += av[ i ]->len;
	}
	v2 = (crunch_vector)crunch_alloc(len * sizeof(crunch_vector_block *));
	v2->len = len;
	v2->flags = v1->flags;
	crunch_vector_block **p = v2->data;
	for(i = 0; i < n; ++i) {
		memcpy(p, av[ i ]->data, av[ i ]->size);
		if((av[ i ]->flags & CRUNCH_NESTEDRC) != 0) {
			for(int j = 0; j < av[ i ]->len; ++j)
				crunch_addref(((crunch_vector_block **)av[ i ]->data)[ j ]);
		}
	}
	return v2;
}

static crunch_string crunch_make_string(int argc, long len, ...) {
	va_list va;
	va_start(va, len);
	uint32_t u = ' ';
	if(argc > 1) u = va_arg(va, int);
#ifndef CRUNCH_NO_UTF
	char buf[ 4 ];
	char *p = crunch_utf_encode(u, buf);
	int c = p - buf;
#else
	int c = 1;
#endif
	crunch_string str = (crunch_string)crunch_alloc(c * len + 1); // plus null byte
	str->len = len;
#ifndef CRUNCH_NO_UTF
	for(p = str->data; len--; p += c) memcpy(p, buf, c);
#else
	memset(str->data, u, len);
#endif
       str->data[ str->size - 1 ] = '\0';
	return str;
}

static crunch_string crunch_string_at(char *p) {
	int n = strlen(p);
	crunch_string s = (crunch_string)crunch_alloc(n + 1);
	memcpy(s->data, p, n);
#ifdef CRUNCH_NO_UTF
	s->len = n;
#else
	s->len = crunch_utf_count(p, n);
#endif
	return s;
}

static crunch_string crunch_string_append(int n, crunch_string s1, ...) {
	va_list va;
	va_start(va, s1);
	crunch_string as[ 10 ];
	as[ 0 ] = s1;
	int i, total = s1->size - 1;
	for(i = 1; i < n; ++i) {
		as[ i ] = va_arg(va, crunch_string);
		total += as[ i ]->size - 1;
	}
	crunch_string v2 = (crunch_string)crunch_alloc(total + 1);
	char *p = v2->data;
	v2->len = 0;
	for(i = 0; i < n; ++i) {
		memcpy(p, as[ i ]->data, as[ i ]->size - 1);
		p += as[ i ]->size - 1;
		v2->len += as[ i ]->len;
	}
	return v2;
}

static inline void *crunch_string_iterator(crunch_string str) { return str->data; }

static inline int crunch_string_iterator_at_end(void *ptr) { return *((char *)ptr) == '\0'; }

static inline uint32_t crunch_string_iterator_ref(void *ptr) {
#ifdef CRUNCH_NO_UTF
	return *((char *)ptr);
#else
	uint32_t c;
	crunch_utf_decode(ptr, &c);
	return c;
#endif
}

static inline void *crunch_string_iterator_next(void *ptr) {
#ifdef CRUNCH_NO_UTF
	return (void *)((char *)ptr + 1);
#else
	uint32_t c;
	return crunch_utf_decode(ptr, &c);
#endif
}

static inline crunch_string crunch_string_buffer(crunch_string s) {
	crunch_string s2;
#ifdef CRUNCH_NO_UTF
	s2 = (crunch_string)crunch_alloc(s->size + 1);
	memset(s2->data, 0, s->size + 1);
	s2->len = s->size;
#else
	s2 = (crunch_string)crunch_alloc(s->size * 4 + 1);
	s2->len = s->len;
	memset(s2->data, 0, s->size * 4 + 1);
#endif
	return s2;
}

static inline void *crunch_string_buffer_add(crunch_string s, void *b, uint32_t c) {
#ifdef CRUNCH_NO_UTF
	++s->size;
	*((char *)b) = c;
	return (void *)((char *)b + 1);
#else
	return crunch_utf_encode(c, (char *)b);
#endif
}

static inline void crunch_string_buffer_finalize(crunch_string s, void *b) {
#ifdef CRUNCH_NO_UTF
	s->len = s->size;
#else
	s->len = s->size = (char *)b - s->data;
#endif
}

static inline char *crunch_string_index(crunch_string s, crunch_word p) {
#ifdef CRUNCH_NO_UTF
	return s->data + p;
#else
	return crunch_utf_index(s, p);
#endif
}

static inline uint32_t crunch_string_ref(crunch_string s, long p) {
	assert(p >= 0 && p < s->len);
#ifdef CRUNCH_NO_UTF
	return s->data[ p ];
#else
	uint32_t c;
	if(s->len == s->size) return s->data[ p ];
	crunch_utf_decode(crunch_utf_index(s, p), &c);
	return c;
#endif
}

static void crunch_string_set(crunch_string s, long p, uint32_t c) {
	assert(p >= 0 && p < s->len);
#ifdef CRUNCH_NO_UTF
	s->data[ p ]  = c;
#else
	if(s->size == s->len && c < 127) s->data[ p ] = c;
	else {
		char buf[ 4 ];
		char *bp = crunch_utf_encode(c, buf);
		char *sp = crunch_utf_index(s, p);
		uint32_t co;
		char *dp = crunch_utf_decode(sp, &co);
		int olen = dp - sp;
		int nlen = bp - buf;
		if(olen == nlen) memcpy(sp, buf, nlen);
		else if(olen > nlen) {
			// string shrinks
			memcpy(sp, buf, nlen);
			memmove(sp + nlen, dp, s->size - (s->data - dp) + 1); // incl. 0 byte
			s->size -= olen - nlen;
		} else {
			// string grows
			int g = nlen - olen;
			s->data = realloc(s->data, s->size + g);
			memmove(dp + g, dp, s->size - (s->data - dp) + 1); // incl. 0 byte
			s->size += g;
		}
	}
#endif
}

static inline crunch_string crunch_substring(crunch_string s, long from, long to) {
	assert(from >= 0 && to >= 0 && to <= s->len);
#ifdef CRUNCH_NO_UTF
	crunch_string s2 = (crunch_string)crunch_alloc(to - from + 1);
	memcpy(s2->data, s->data + from, to - from);
#else
	crunch_string s2 = (crunch_string)crunch_alloc((to - from) * 4 + 1);
	char *p1 = crunch_string_index(s, from);
	char *p2 = crunch_string_index(s, to);
	memcpy(s2->data, p1, p2 - p1);
	int tlen = p2 - p1 + 1;
	if(tlen < s2->size) {
		s2->data = realloc(s2->data, tlen);
		s2->size = tlen;
	}
#endif
	s2->len = to - from;
	return s2;
}

static crunch_string crunch_string_copy(int n, crunch_string s, ...) {
	va_list va;
	va_start(va, s);
	int start = 0, end = s->len;
	if(n > 2) start = va_arg(va, long);
	if(n > 3) end = va_arg(va, long);
	int len = end - start;
	assert(start >= 0 && start < s->len && start < end);
	assert(end >= 0 && end <= s->len);
#ifdef CRUNCH_NO_UTF
	crunch_string s2 = (crunch_string)crunch_alloc(len);
	memcpy(s2->data, s->data + start, len);
#else
	crunch_string s2 = (crunch_string)crunch_alloc(len * 4);
	char *p = crunch_utf_index(s, start);
	char *pe = crunch_utf_index(s, end);
	memcpy(s2->data, p, pe - p);
#endif
	s2->len = len;
	return s2;
}

static void crunch_string_copy_into(int n, crunch_string to, long at, crunch_string from, ...) {
	va_list va;
	va_start(va, from);
	int start = 0, end = from->len;
	if(n > 3) start = va_arg(va, long);
	if(n > 4) end = va_arg(va, long);
	assert(start >=0 && start < from->len && start < end);
	assert(end >= 0 && end <= from->len);
	int len = end - start;
	assert(at + len < to->len);
#ifdef CRUNCH_NO_UTF
	memcpy(to->data + at, from->data + start, len);
#else
	char *pfs = crunch_utf_index(from, start);
	char *pfe = crunch_utf_index(from, end);
	char *pts = crunch_utf_index(to, at);
	char *pte = crunch_utf_index(to, at + len);
	int flen = pfe - pfs, tlen = pte - pts;
	if(flen > tlen) {
		int diff = flen - tlen;
		int p = pts - to->data;
		to->data = realloc(to->data, to->size + diff);
		assert(to->data != NULL);
		pts = to->data + p;
		memmove(pts + tlen + diff, pts + tlen, diff + 1);
	} else if(flen < tlen) {
		int diff = tlen - flen;
		memmove(pts + diff, pte, diff + 1);
		to->data = realloc(to->data, to->size - diff);
		assert(to->data != NULL);
	}
	memmove(pts, pfs, flen);
#endif
}

static void crunch_string_fill(int n, crunch_string s, uint32_t c, ...) {
	va_list va;
	va_start(va, c);
	int start = 0, end = s->len;
	if(n > 2) start = va_arg(va, long);
	if(n > 3) end = va_arg(va, long);
	assert(start >= 0 && start < s->len && start < end);
	assert(end >= 0 && end <= s->len);
	int len = end - start;
#ifdef CRUNCH_NO_UTF
	memset(s->data + start, c, len);
#else
	char buf[ 4 ];
	char *e = crunch_utf_encode(c, buf);
	int flen = e - buf;
	char *ps = crunch_utf_index(s, start);
	char *pe = crunch_utf_index(s, end);
	int tlen = pe - ps;
	if(tlen < flen) {
		int diff = tlen - flen;
		int p = ps - s->data;
		s->data = realloc(s->data, s->size + diff);
		assert(s->data != NULL);
		ps = s->data + p;
		memmove(ps + flen + diff, ps + flen, diff + 1);
	} else if(flen < tlen) {
		int diff = tlen - flen;
		memmove(pe - diff, pe, diff + 1);
		s->data = realloc(s->data, s->size - diff);
		assert(s->data != NULL);
	}
	for(int i = 0; i < len; ++i) {
		memmove(ps, buf, flen);
		ps += flen;
	}
#endif
}

static inline long crunch_exact_float(double n) {
	double t = trunc(n);
	if(n == t) return t;
	crunch_abort("can not convert float to exact number");
	return 0;
}

static inline long crunch_exact_complex(long n) {
	double t = trunc(creal(n));
	if(creal(n) == t && cimag(n) == 0) return t;
	crunch_abort("can not convert complex to exact number");
	return 0;
}

static inline double crunch_inexact_integer(long n) {
	return (double)n;
}

static crunch_charvector crunch_string_to_vector(int n, crunch_string s, ...) {
	va_list va;
	va_start(va, s);
	int from = 0, to = s->len;
	if(n > 1) from = va_arg(va, long);
	if(n > 2) to = va_arg(va, long);
	assert(from >= 0 && from < s->len && to >= 0 && to < s->len && to >= from);
	crunch_charvector v = (crunch_charvector)crunch_alloc(to - from);
#ifdef CRUNCH_NO_UTF
	memcpy(v->data, s->data, to - from);
#else
	char *p = s->data;
	uint32_t *p2 = v->data;
	while(from < to) p = crunch_utf_decode(p, p2++);
#endif
	v->len = to - from;
	return v;
}

static crunch_string crunch_vector_to_string(int n, crunch_charvector v, ...) {
	va_list va;
	va_start(va, v);
	int from = 0, to = v->len;
	if(n > 1) from = va_arg(va, long);
	if(n > 2) to = va_arg(va, long);
	assert(from >= 0 && from < v->len && to >= 0 && to < v->len && to >= from);
#ifdef CRUNCH_NO_UTF
	crunch_string s = (crunch_string)crunch_alloc(to - from);
	memcpy(s->data, v->data, to - from);
#else
	crunch_string s = (crunch_string)crunch_alloc((to - from) * 4 + 1);
	uint32_t *p = v->data;
	char *p2 = s->data;
	while(from < to) p2 = crunch_utf_encode(*(p++), p2);
	*(p2++) = '\0';
	int tlen = p2 - s->data;
	if(tlen < s->size) {
		s->data = realloc(s->data, tlen + 1);
		s->size = tlen;
	}
#endif
	s->len = to - from;
	return s;
}

static inline int crunch_even(long n) { return (n & 1) == 0; }
static inline int crunch_odd(long n) { return (n & 1) != 0; }

static inline int crunch_binary_port(crunch_port p) {
	return (p->flags & CRUNCH_TEXTPORT) == 0;
}

static inline int crunch_textual_port(crunch_port p) {
	return (p->flags & CRUNCH_TEXTPORT) != 0;
}

static inline int crunch_input_port(crunch_port p) {
	return (p->flags & CRUNCH_INPUTPORT) != 0;
}

static inline int crunch_output_port(crunch_port p) {
	return (p->flags & CRUNCH_OUTPUTPORT) != 0;
}

static void crunch_flush_output_port(int n, ...) {
	va_list va;
	va_start(va, n);
	crunch_port port = crunch_stdout;
	crunch_ensure_stdio();
	if(n > 1) port = va_arg(va, crunch_port);
	fflush(port->data->file);
}

static crunch_string crunch_integer_to_string(int n, long x, ...) {
	va_list va;
	va_start(va, x);
	int base = 10;
	if(n > 1) base = va_arg(va, long);
	char buf[ 64 ], *p = buf + 64;
	*(p--) = '\0';
	int sg = x < 0;
	x = labs(x);
	while(x > 0) {
		*(p--) = x % base;
		x /= base;
	}
	if(sg) *p = '-';
	crunch_string s = (crunch_string)crunch_alloc(strlen(p) + 1);
	s->len = s->size - 1;
	memcpy(s->data, p, s->size);
	return s;
}

#define CRUNCH_FLOAT_PRECISION	20

static crunch_string crunch_float_to_string(int n, double x, ...) {
	char buf[ 64 ];
	snprintf(buf, sizeof(buf), "%.*g", CRUNCH_FLOAT_PRECISION, x);
	if(strchr(buf, '\0') == NULL) strcat(buf, ".0");
	crunch_string s = (crunch_string)crunch_alloc(strlen(buf) + 1);
	s->len = s->size - 1;
	memcpy(s->data, buf, s->size);
	return s;
}

static crunch_string crunch_complex_to_string(int n, double complex x, ...) {
	char buf[ 64 ];
	snprintf(buf, sizeof(buf), "%.*g+i%.*g",CRUNCH_FLOAT_PRECISION, creal(x),
		CRUNCH_FLOAT_PRECISION, cimag(x));
	crunch_string s = (crunch_string)crunch_alloc(strlen(buf) + 1);
	s->len = s->size - 1;
	memcpy(s->data, buf, s->size);
	return s;
}

static crunch_string crunch_utf8_to_string(int n, crunch_u8vector v, ...) {
	va_list va;
	va_start(va, v);
	int start = 0, end = v->len;
	if(n > 1) start = va_arg(va, long);
	if(n > 2) end = va_arg(va, long);
	assert(start >= 0 && start <= v->len && start <= end);
	assert(end >= 0 && end <= v->len);
	char *p = (char *)v->data;
	crunch_string s = (crunch_string)crunch_alloc(end - start + 1);
	memcpy(s->data, v->data + start, end - start);
#ifdef CRUNCH_NO_UTF
	s->len = s->size - 1;
#else
	s->len = crunch_utf_count(s->data, s->size - 1);
#endif
	return s;
}

static crunch_u8vector crunch_string_to_utf8(int n, crunch_string s, ...) {
	va_list va;
	va_start(va, s);
	int start = 0, end = s->len;
	if(n > 1) start = va_arg(va, long);
	if(n > 2) end = va_arg(va, long);
	assert(start >= 0 && start <= s->len && start <= end);
	assert(end >= 0 && end <= s->len);
#ifdef CRUNCH_NO_UTF
	crunch_u8vector v = (crunch_u8vector)crunch_alloc(end - start);
	memcpy(v->data, s->data + start, end - start);
#else
	char *ps = crunch_utf_index(s, start);
	char *pe = crunch_utf_index(s, end);
	crunch_u8vector v = (crunch_u8vector)crunch_alloc(pe - ps);
	memcpy(v->data, ps, pe - ps);
#endif
	v->len = v->size;
	return v;
}

static inline void crunch_close_port(crunch_port p) {
	if((p->flags & CRUNCH_OPEN) != 0) fclose(p->data->file);
	p->flags = p->flags & ~CRUNCH_OPEN;
}

static inline crunch_port crunch_current_input_port(int n, ...) {
	va_list va;
	va_start(va, n);
	if(n > 0) crunch_stdin = va_arg(va, crunch_port);
	crunch_ensure_stdio();
	return crunch_stdin;
}

static inline crunch_port crunch_current_output_port(int n, ...) {
	va_list va;
	va_start(va, n);
	if(n > 0) crunch_stdout = va_arg(va, crunch_port);
	crunch_ensure_stdio();
	return crunch_stdout;
}

static inline crunch_port crunch_current_error_port(int n, ...) {
	va_list va;
	va_start(va, n);
	if(n > 0) crunch_stderr = va_arg(va, crunch_port);
	crunch_ensure_stdio();
	return crunch_stderr;
}

static inline int crunch_port_open(crunch_port p) {
	return (p->flags & CRUNCH_OPEN) != 0;
}

static crunch_port crunch_open_input_string(crunch_string s) {
	crunch_port p = (crunch_port)crunch_alloc(sizeof(crunch_port_info));
	p->flags |= CRUNCH_PORT | CRUNCH_TEXTPORT | CRUNCH_INPUTPORT |
		CRUNCH_REFPORT | CRUNCH_OPEN;
	p->data->file = fmemopen(s->data, s->size, "r");
	if(p->data->file == NULL) p->flags &= ~CRUNCH_OPEN;
	p->data->ref = (crunch_vector_block*)s;
	p->len = p->size;
	crunch_addref(s);
	return p;
}

static crunch_port crunch_open_input_bytevector(crunch_u8vector s) {
	crunch_port p = (crunch_port)crunch_alloc(sizeof(crunch_port_info));
	p->flags |= CRUNCH_PORT | CRUNCH_INPUTPORT | CRUNCH_REFPORT |
		CRUNCH_OPEN;
	p->data->file = fmemopen(s->data, s->size, "rb");
	if(p->data->file == NULL) p->flags &= ~CRUNCH_OPEN;
	p->data->ref = (crunch_vector_block *)s;
	p->len = p->size;
	crunch_addref(s);
	return p;
}

static crunch_port crunch_open_output_string(void) {
	crunch_port p = (crunch_port)crunch_alloc(sizeof(crunch_port_info));
	p->flags |= CRUNCH_PORT | CRUNCH_TEXTPORT | CRUNCH_OUTPUTPORT |
		CRUNCH_REFPORT | CRUNCH_OPEN;
	crunch_u8vector buf = (crunch_u8vector)crunch_alloc(CRUNCH_BUFFER_SIZE);
	CRUNCH_INCRC(buf->rc);
	p->data->file = fmemopen(buf->data, CRUNCH_BUFFER_SIZE, "w");
	if(p->data->file == NULL) p->flags &= ~CRUNCH_OPEN;
	p->data->ref = (crunch_vector_block *)buf;
	p->len = p->size;
	return p;
}

static crunch_port crunch_open_output_bytevector(crunch_u8vector s) {
	crunch_port p = (crunch_port)crunch_alloc(sizeof(crunch_port_info));
	p->flags |= CRUNCH_PORT | CRUNCH_OUTPUTPORT | CRUNCH_REFPORT |
		CRUNCH_OPEN;
	crunch_u8vector buf = (crunch_u8vector)crunch_alloc(CRUNCH_BUFFER_SIZE);
	CRUNCH_INCRC(buf->rc);
	p->data->file = fmemopen(buf->data, CRUNCH_BUFFER_SIZE, "wb");
	if(p->data->file == NULL) p->flags &= ~CRUNCH_OPEN;
	p->data->ref = (crunch_vector_block *)buf;
	p->len = p->size;
	return p;
}

static crunch_string crunch_get_output_string(crunch_port p) {
	if((p->flags & CRUNCH_TEXTPORT) == 0 || (p->flags & CRUNCH_OUTPUTPORT) == 0 ||
		(p->flags & CRUNCH_REFPORT) == 0)
		crunch_abort("get-output-string: not an output string port");
	long pos = ftell(p->data->file);
	if(p < 0) crunch_abort(strerror(errno));
	crunch_string s = (crunch_string)crunch_alloc(pos + 1);
	memcpy(s->data, p->data->ref->data, pos);
#ifdef CRUNCH_NO_UTF
	s->len = pos;
#else
	s->len = crunch_utf_count(s->data, pos);
#endif
	return s;
}

static crunch_u8vector crunch_get_output_bytevector(crunch_port p) {
	if((p->flags & CRUNCH_TEXTPORT) != 0 || (p->flags & CRUNCH_OUTPUTPORT) == 0 ||
		(p->flags & CRUNCH_REFPORT) == 0)
		crunch_abort("get-output-bytevector: not an output bytevector port");
	long pos = ftell(p->data->file);
	if(p < 0) crunch_abort(strerror(errno));
	crunch_u8vector s = (crunch_u8vector)crunch_alloc(pos);
	memcpy(s->data, p->data->ref->data, pos);
	s->len = pos;
	return s;
}

static void crunch_write_u8(int n, long u, ...) {
	va_list va;
	va_start(va, u);
	crunch_port p = crunch_stdout;
	crunch_ensure_stdio();
	if(n > 1) p = va_arg(va, crunch_port);
	fputc(u, p->data->file);
}

static void crunch_write_char(int n, uint32_t c, ...) {
	va_list va;
	va_start(va, c);
	crunch_port p = crunch_stdout;
	crunch_ensure_stdio();
	if(n > 1) p = va_arg(va, crunch_port);
#ifdef CRUNCH_NO_UTF
	fputc(c, p->data->file);
#else
	char buf[ 4 ];
	char *e = crunch_utf_encode(c, buf);
	fwrite(buf, 1, e - buf, p->data->file);
#endif
}

static void crunch_write_bytevector(int n, crunch_u8vector v, ...) {
	va_list va;
	va_start(va, v);
	crunch_port p = crunch_stdout;
	crunch_ensure_stdio();
	int start = 0, end = v->len;
	if(n > 1) p = va_arg(va, crunch_port);
	if(n > 2) start = va_arg(va, long);
	if(n > 3) end = va_arg(va, long);
	assert(start >= 0 && start <= v->len && start <= end);
	assert(end >= 0 && end <= v->len);
	fwrite(v->data, 1, end - start, p->data->file);
}

static void crunch_write_string(int n, crunch_string s, ...) {
	va_list va;
	va_start(va, s);
	crunch_port p = crunch_stdout;
	crunch_ensure_stdio();
	int start = 0, end = s->len;
	if(n > 1) p = va_arg(va, crunch_port);
	if(n > 2) start = va_arg(va, long);
	if(n > 3) end = va_arg(va, long);
	assert(start >= 0 && start <= s->len && start <= end);
	assert(end >= 0 && end <= s->len);
#ifdef CRUNCH_NO_UTF
	fwrite(s->data + start, 1, end - start, p->data->file);
#else
	char *ps = crunch_utf_index(s, start);
	char *pe = crunch_utf_index(s, end);
	fwrite(ps, 1, pe - ps, p->data->file);
#endif
}

static int crunch_char_ready(int n, ...) {
	va_list va;
	va_start(va, n);
	crunch_port p = crunch_stdin;
	crunch_ensure_stdio();
	if(n > 0) p = va_arg(va, crunch_port);
	assert((p->flags & CRUNCH_TEXTPORT) != 0);
	return p->data->peeked == -1 && !feof(p->data->file);	//XXX
}

static int crunch_u8_ready(int n, ...) {
	va_list va;
	va_start(va, n);
	crunch_port p = crunch_stdin;
	crunch_ensure_stdio();
	if(n > 0) p = va_arg(va, crunch_port);
	assert((p->flags & CRUNCH_TEXTPORT) == 0);
	return p->data->peeked == -1 && !feof(p->data->file);	//XXX
}

static long crunch_read_u8(int n, ...) {
	va_list va;
	va_start(va, n);
	crunch_port p = crunch_stdin;
	crunch_ensure_stdio();
	if(n > 0) p = va_arg(va, crunch_port);
	assert((p->flags & CRUNCH_TEXTPORT) == 0);
	long r;
	if(p->data->peeked != -1) {
	 	r = p->data->peeked;
		p->data->peeked = -1;
		return r;
	}
	r = fgetc(p->data->file);
	if(r == EOF) return crunch_eof_scalar;
	return r;
}

static long crunch_read_char(int n, ...) {
	va_list va;
	va_start(va, n);
	crunch_port p = crunch_stdin;
	crunch_ensure_stdio();
	if(n > 0) p = va_arg(va, crunch_port);
	assert((p->flags & CRUNCH_TEXTPORT) != 0);
	long r;
	if(p->data->peeked != -1) {
		r = p->data->peeked;
		p->data->peeked = -1;
		return r;
	}
	r = fgetc(p->data->file);
	if(r == EOF) return crunch_eof_scalar;
	return r;
}

static long crunch_peek_char(int n, ...) {
	va_list va;
	va_start(va, n);
	crunch_port p = crunch_stdin;
	crunch_ensure_stdio();
	if(n > 0) p = va_arg(va, crunch_port);
	assert((p->flags & CRUNCH_TEXTPORT) != 0);
	long r;
	if(p->data->peeked != -1) return p->data->peeked;
	r = fgetc(p->data->file);
#ifndef CRUNCH_NO_UTF
	if((r & 0xc0) == 0xc0) {
		char buf[ 4 ], *pb = buf;
		buf[ 0 ] = r;
		do {
			*pb = fgetc(p->data->file);
			if(r == EOF) break;
		} while((*(pb++) & 0xc0) == 0x80);
		uint32_t c;
		crunch_utf_decode(buf, &c);
		return c;
	}
#endif
	if(r == EOF) return crunch_eof_scalar;
	p->data->peeked = r;
	return r;
}

static long crunch_peek_u8(int n, ...) {
	va_list va;
	va_start(va, n);
	crunch_port p = crunch_stdin;
	crunch_ensure_stdio();
	if(n > 0) p = va_arg(va, crunch_port);
	assert((p->flags & CRUNCH_TEXTPORT) == 0);
	long r;
	if(p->data->peeked != -1) return p->data->peeked;
	r = fgetc(p->data->file);
	if(r == EOF) return crunch_eof_scalar;
	p->data->peeked = r;
	return r;
}

static crunch_string crunch_read_string(int n, long num, ...) {
	va_list va;
	va_start(va, num);
	crunch_port p = crunch_stdin;
	crunch_ensure_stdio();
	if(n > 1) p = va_arg(va, crunch_port);
	assert((p->flags & CRUNCH_TEXTPORT) != 0);
	crunch_string buf = (crunch_string)crunch_alloc(num * 4 + 1);
	int total = 0;
	char *pb = buf->data;
	if(p->data->peeked != -1) {
#ifndef CRUNCH_NO_UTF
		pb = crunch_utf_encode(p->data->peeked, pb);
#else
		*(pb++) = p->data->peeked;
#endif
		p->data->peeked = -1;
		--num;
		++total;
	}
	while(num > 0) {
		int r = fread(pb, 1, num, p->data->file);
		uint32_t c;
		if(r == 0) break;
		while(r--) {
			pb = crunch_utf_decode(pb, &c);
			++total;
			--num;
		}
	}
	if(total == 0) {
		crunch_release((crunch_vector_block *)buf);
		return (crunch_string)&crunch_eof_vector;
	}
	buf->data = realloc(buf->data, pb - buf->data + 1);
	return buf;
}

static crunch_u8vector crunch_read_bytevector(int n, long num, ...) {
	va_list va;
	va_start(va, num);
	crunch_port p = crunch_stdin;
	crunch_ensure_stdio();
	if(n > 1) p = va_arg(va, crunch_port);
	assert((p->flags & CRUNCH_TEXTPORT) == 0);
	crunch_u8vector buf = (crunch_u8vector)crunch_alloc(num);
	int total = 0;
	uint8_t *pb = buf->data;
	if(p->data->peeked != -1) {
		*(pb++) = p->data->peeked;
		p->data->peeked = -1;
		--num;
		++total;
	}
	while(num > 0) {
		int r = fread(pb, 1, num, p->data->file);
		if(r == 0) break;
		total += r;
		num -= r;
	}
	if(total == 0) {
		crunch_release((crunch_vector_block *)buf);
		return (crunch_u8vector)&crunch_eof_vector;
	}
	buf->data = realloc(buf->data, pb - buf->data + 1);
	return buf;
}

static long crunch_read_bytevector_into(int n, crunch_u8vector v, ...) {
	va_list va;
	va_start(va, v);
	crunch_port p = crunch_stdin;
	crunch_ensure_stdio();
	int start = 0, end = v->len;
	if(n > 1) p = va_arg(va, crunch_port);
	if(n > 2) start = va_arg(va, long);
	if(n > 3) end = va_arg(va, long);
	assert(start >= 0 && start <= v->len && start <= end);
	assert(end >= 0 && end <= v->len);
	assert((p->flags & CRUNCH_TEXTPORT) == 0);
	int total = 0;
	char *pb = (char *)v->data + start;
	if(p->data->peeked != -1) {
		*(pb++) = p->data->peeked;
		p->data->peeked = -1;
		++start;
		++total;
	}
	while(start < end) {
		int r = fread(pb, 1, end - start, p->data->file);
		if(r == 0) break;
		start += r;
		++total;
	}
	return total;
}


/* (scheme file) */

static crunch_port crunch_open_input_file(crunch_string s) {
	crunch_port p = (crunch_port)crunch_alloc(sizeof(crunch_port_info));
	p->flags |= CRUNCH_PORT | CRUNCH_TEXTPORT | CRUNCH_INPUTPORT |
		CRUNCH_OPEN;
	p->data->file = fopen(s->data, "r");
	if(p->data->file == NULL) p->flags &= ~CRUNCH_OPEN;
	p->data->peeked = -1;
	return p;
}

static crunch_port crunch_open_output_file(crunch_string s) {
	crunch_port p = (crunch_port)crunch_alloc(sizeof(crunch_port_info));
	p->flags |= CRUNCH_PORT | CRUNCH_TEXTPORT | CRUNCH_OUTPUTPORT |
		CRUNCH_OPEN;
	p->data->file = fopen(s->data, "w");
	if(p->data->file == NULL) p->flags &= ~CRUNCH_OPEN;
	p->data->peeked = -1;
	return p;
}

static crunch_port crunch_open_binary_input_file(crunch_string s) {
	crunch_port p = (crunch_port)crunch_alloc(sizeof(crunch_port_info));
	p->flags |= CRUNCH_PORT | CRUNCH_INPUTPORT | CRUNCH_OPEN;
	p->data->file = fopen(s->data, "rb");
	if(p->data->file == NULL) p->flags &= ~CRUNCH_OPEN;
	p->data->peeked = -1;
	return p;
}

static crunch_port crunch_open_binary_output_file(crunch_string s) {
	crunch_port p = (crunch_port)crunch_alloc(sizeof(crunch_port_info));
	p->flags |= CRUNCH_PORT | CRUNCH_OUTPUTPORT | CRUNCH_OPEN;
	p->data->file = fopen(s->data, "wb");
	if(p->data->file == NULL) p->flags &= ~CRUNCH_OPEN;
	p->data->peeked = -1;
	return p;
}

static inline void crunch_delete_file(crunch_string s) {
	unlink(s->data);
}

static inline int crunch_file_exists(crunch_string s) {
	struct stat st;
	if(stat(s->data, &st) == -1 && errno == ENOENT) return 0;
	return 1;
}


/* (scheme write) */

static inline void crunch_write_integer(int argc, long n, ...) {
	va_list va;
	va_start(va, n);
	crunch_port port = crunch_stdout;
	crunch_ensure_stdio();
	if(argc > 1) port = va_arg(va, crunch_port);
	fprintf(port->data->file, "%ld", n);
}

static inline void crunch_write_float(int argc, double n, ...) {
	va_list va;
	char buf[ 64 ];
	va_start(va, n);
	crunch_port port = crunch_stdout;
	crunch_ensure_stdio();
	if(argc > 1) port = va_arg(va, crunch_port);
	snprintf(buf, sizeof(buf), "%g", n);
       fputs(buf, port->data->file);
       if(!strchr(buf, '.')) fputs(".0", port->data->file);
}

static inline void crunch_write_complex(int argc, double complex n, ...) {
	va_list va;
	va_start(va, n);
	crunch_port port = crunch_stdout;
	crunch_ensure_stdio();
	if(argc > 1) port = va_arg(va, crunch_port);
	fprintf(port->data->file, "%g + %gi", creal(n), cimag(n));
}

static void crunch_write_pointer(int argc, void *n, ...) {
	va_list va;
	va_start(va, n);
	crunch_port port = crunch_stdout;
	crunch_ensure_stdio();
	if(argc > 1) port = va_arg(va, crunch_port );
	fprintf(port->data->file, "<%p>", n);
}

static inline void crunch_write_anon(int argc, char *str, ...) {
	va_list va;
	va_start(va, str);
	crunch_port port = crunch_stdout;
	crunch_ensure_stdio();
	if(argc > 1) port = va_arg(va, crunch_port);
	fputs(str, port->data->file);
}

#define crunch_write_struct(argc, x, ...)	crunch_write_anon(argc, "<struct>", __VA_ARGS__)

static void crunch_write_char_readable(int argc, int n, ...) {
	va_list va;
	va_start(va, n);
	crunch_port port = crunch_stdout;
	crunch_ensure_stdio();
	if(argc > 1) port = va_arg(va, crunch_port);
	if(n < 33 || n > 126) fprintf(port->data->file, "#\\x%x;", n);
	else fprintf(port->data->file, "#\\%c", n);
}

static void crunch_write_boolean(int argc, int n, ...) {
	va_list va;
	va_start(va, n);
	crunch_port port = crunch_stdout;
	crunch_ensure_stdio();
	if(argc > 1) port = va_arg(va, crunch_port);
	fprintf(port->data->file, "#%c", (n ? 't' : 'f'));
}

static void crunch_write_string_readable(int argc, crunch_string n, ...) {
	va_list va;
	va_start(va, n);
	crunch_port port = crunch_stdout;
	crunch_ensure_stdio();
	if(argc > 1) port = va_arg(va, crunch_port );
	int len = n->len;
       char *p = n->data;
       fputc('"', port->data->file);
       while(len--) {
		uint32_t c;
		int e;
#ifdef CRUNCH_NO_UTF
		c = *(p++);
#else
   	    	p = crunch_utf_decode(p, &c);
#endif
   	    	if(c < 32 || c > 126) fprintf(port->data->file, "\\x%d;", c);
   	    	else fputc(c, port->data->file);
       }
       fputc('"', port->data->file);
}

#define crunch_write_numvector(v, t, pr, fmt) \
static void crunch_write_ ## v(int argc, crunch_ ## v n, ...) { \
	va_list va; \
	va_start(va, n); \
	crunch_port port = crunch_stdout; \
	crunch_ensure_stdio(); \
	if(argc > 1) port = va_arg(va, crunch_port ); \
	int len = n->len; \
       t *p = n->data; \
       fputs( "#" pr "(", port->data->file); \
       if(len > 0) { \
	       fprintf(port->data->file, fmt, *(p++)); \
 	      while(--len) fprintf(port->data->file, " " fmt, *(p++)); \
 	} \
       fputc(')', port->data->file); \
}

crunch_write_numvector(u8vector, uint8_t, "u8", "%" PRIu8)
crunch_write_numvector(s8vector, int8_t, "s8", "%" PRIi8)
crunch_write_numvector(u16vector, uint16_t, "u16", "%" PRIu16)
crunch_write_numvector(s16vector, int16_t, "s16", "%" PRIi16)
crunch_write_numvector(u32vector, uint32_t, "u32", "%" PRIu32)
crunch_write_numvector(s32vector, int32_t, "s32", "%" PRIi32)
crunch_write_numvector(u64vector, uint64_t, "u64", "%" PRIu64)
crunch_write_numvector(s64vector, int64_t, "s64", "%" PRIi64)
crunch_write_numvector(f32vector, float, "f32", "%g")
crunch_write_numvector(f64vector, double, "f64", "%g")

static void crunch_write_charvector(int argc, crunch_u32vector n, ...) {
	va_list va;
	va_start(va, n);
	crunch_port port = crunch_stdout;
	crunch_ensure_stdio();
	if(argc > 1) port = va_arg(va, crunch_port );
	int len = n->len;
       fputs("#(", port->data->file);
       int f = 0;
       uint32_t *p = n->data;
       char buf[ 5 ];
     	while(--len) {
     		if(f) fputc(' ', port->data->file);
     		f = 1;
     		fputs("#\\", port->data->file);
     		uint32_t c = *(p++);
		if(c > 32 && c < 128) fputc(c, port->data->file);
		else fprintf(port->data->file, "x%x", c);
      	}
       fputc(')', port->data->file);
}

static void crunch_write_c64vector(int argc, crunch_c64vector n, ...) {
	va_list va;
	va_start(va, n);
	crunch_port port = crunch_stdout;
	crunch_ensure_stdio();
	if(argc > 1) port = va_arg(va, crunch_port );
	int len = n->len;
       float complex *p = n->data;
       fputs("#c64(", port->data->file);
       if(len > 0) {
		fprintf(port->data->file, "%g+%gi", creal(*p), cimag(*p));
 	      	while(--len) {
 	      		++p;
  	    		fprintf(port->data->file, " %g+%gi", creal(*p), cimag(*p));
   	   	}
      	}
       fputc(')', port->data->file);
}

static void crunch_write_c128vector(int argc, crunch_c128vector n, ...) {
	va_list va;
	va_start(va, n);
	crunch_port port = crunch_stdout;
	crunch_ensure_stdio();
	if(argc > 1) port = va_arg(va, crunch_port );
	int len = n->len;
       double complex *p = n->data;
       fputs("#c128(", port->data->file);
       if(len > 0) {
	       fprintf(port->data->file, "%g+%gi", creal(*p), cimag(*p));
 	      while(--len) {
 	      		++p;
  	    		fprintf(port->data->file, " %g+%gi", creal(*p), cimag(*p));
   	      }
      	}
       fputc(')', port->data->file);
}

static void crunch_write_ptrvector(int argc, crunch_vector n, ...) {
	va_list va;
	va_start(va, n);
	crunch_port port = crunch_stdout;
	crunch_ensure_stdio();
	if(argc > 1) port = va_arg(va, crunch_port );
	int len = n->len;
       void **p = (void **)n->data;
       fputs("#(", port->data->file);
       if(len > 0) {
	       fprintf(port->data->file, "%p", *(p++));
	       while(--len) fprintf(port->data->file, " %p", *(p++));
	}
       fputc(')', port->data->file);
}

static void crunch_write_stringvector(int argc, crunch_vector n, ...) {
	va_list va;
	va_start(va, n);
	crunch_port port = crunch_stdout;
	crunch_ensure_stdio();
	if(argc > 1) port = va_arg(va, crunch_port);
	int len = n->len;
       crunch_string *p = (crunch_string *)n->data;
       fputs("#(", port->data->file);
       if(len > 0) {
	       crunch_write_string_readable(2, *(p++), port);
	       while(--len) {
	       		fputc(' ', port->data->file);
	       		crunch_write_string_readable(2, *(p++), port);
	        }
	}
       fputc(')', port->data->file);
}

static void crunch_write_boolvector(int argc, crunch_u8vector n, ...) {
	va_list va;
	va_start(va, n);
	crunch_port port = crunch_stdout;
	crunch_ensure_stdio();
	if(argc > 1) port = va_arg(va, crunch_port);
	int len = n->len;
       uint8_t *p = n->data;
       fputs("#(", port->data->file);
       if(len > 0) {
	       crunch_write_boolean(2, *(p++), port);
	       while(--len) {
	       		fputc(' ', port->data->file);
	       		crunch_write_boolean(2, *(p++), port);
	        }
	}
       fputc(')', port->data->file);
}

static void crunch_display_char(int argc, uint32_t n, ...) {
	va_list va;
	va_start(va, n);
	crunch_port port = crunch_stdout;
	crunch_ensure_stdio();
	if(argc > 1) port = va_arg(va, crunch_port);
#ifdef CRUNCH_NO_UTF
	fprintf(port->data->file, "%c", n);
#else
	char buf[ 5 ];
	char *p = crunch_utf_encode(n, buf);
	fprintf(port->data->file, "%.*s", (int)(p - buf), buf);
#endif
}

static void crunch_display_string(int argc, crunch_string n, ...) {
	va_list va;
	va_start(va, n);
	crunch_port port = crunch_stdout;
	crunch_ensure_stdio();
	if(argc > 1) port = va_arg(va, crunch_port);
	fputs(n->data, port->data->file);
}

static void crunch_display_stringvector(int argc, crunch_vector n, ...) {
	va_list va;
	va_start(va, n);
	crunch_port port = crunch_stdout;
	if(argc > 1) port = va_arg(va, crunch_port);
	int len = n->len;
       crunch_string *p = (crunch_string *)n->data;
       fputs("#(", port->data->file);
       if(len > 0) {
	       crunch_display_string(2, *(p++), port);
	       while(--len) {
	       		fputc(' ', port->data->file);
	       		crunch_display_string(2, *(p++), port);
	        }
	}
       fputc(')', port->data->file);
}


/* (scheme complex) */

#define crunch_make_rectangular(x, y) ((x) + I * (y))

static inline double complex crunch_make_polar(double x, double y) {
	return x * cos(y) + I * x * sin(y);
}

#define crunch_float_angle(x) (atan2(0, (x)))


/* (scheme inexact) */

static inline int crunch_complex_finitep(double complex n) {
	return isfinite(creal(n)) && isfinite(cimag(n));
}

static inline int crunch_complex_infinitep(double complex n) {
	return isinf(creal(n)) || isinf(cimag(n));
}

static inline int crunch_complex_nanp(double complex n) {
	return isnan(creal(n)) || isnan(cimag(n));
}

#define crunch_atan(n, x)	atan(x)
#define crunch_atan2(n, x, y)	atan2((x), (y))
#define crunch_catan(n, x)	catan(x)


/* (scheme time) */

static inline long crunch_current_jiffy(void) {
	struct timespec tv;
	clock_gettime(CLOCK_MONOTONIC, &tv);
	return tv.tv_sec * 1000 + tv.tv_nsec / 1000000;
}

static inline long crunch_jiffies_per_second(void) {
	return 1000;
}

static inline long crunch_current_second(void) {
	struct timeval tv;
	gettimeofday(&tv, NULL);
	return tv.tv_sec;
}


/* (scheme char) */

static inline int crunch_char_ci_equal(uint32_t x, uint32_t y) {
	return crunch_utf_char_foldcase(x) == crunch_utf_char_foldcase(y);
}

static inline int crunch_char_ci_greater(uint32_t x, uint32_t y) {
	return crunch_utf_char_foldcase(x) > crunch_utf_char_foldcase(y);
}

static inline int crunch_char_ci_less(uint32_t x, uint32_t y) {
	return crunch_utf_char_foldcase(x) == crunch_utf_char_foldcase(y);
}

static inline int crunch_char_ci_greater_or_equal(uint32_t x, uint32_t y) {
	return crunch_utf_char_foldcase(x) >= crunch_utf_char_foldcase(y);
}

static inline int crunch_char_ci_less_or_equal(uint32_t x, uint32_t y) {
	return crunch_utf_char_foldcase(x) <= crunch_utf_char_foldcase(y);
}

static inline int crunch_string_equal_ci(crunch_string s1, crunch_string s2) {
	return crunch_string_compare_ci(s1, s2) == 0;
}

static inline int crunch_string_less_ci(crunch_string s1, crunch_string s2) {
	int r = crunch_string_compare_ci(s1, s2);
	if(r == 0) return s1->size < s2->size;
	return r < 0;
}

static inline int crunch_string_greater_ci(crunch_string s1, crunch_string s2) {
	int r = crunch_string_compare_ci(s1, s2);
	if(r == 0) return s1->size > s2->size;
	return r > 0;
}

static inline int crunch_string_less_or_equal_ci(crunch_string s1, crunch_string s2) {
	int r = crunch_string_compare_ci(s1, s2);
	if(r == 0) return s1->size <= s2->size;
	return r <= 0;
}

static inline int crunch_string_greater_or_equal_ci(crunch_string s1, crunch_string s2) {
	int r = crunch_string_compare_ci(s1, s2);
	if(r == 0) return s1->size >= s2->size;
	return r >= 0;
}

static crunch_string crunch_string_upcase(crunch_string s) {
#ifdef CRUNCH_NO_UTF
	crunch_string s2 = (crunch_string)crunch_alloc(s->size);
	for(int i = 0; i < s->size; ++i) {
		uint32_t c = s->data[ i ];
		s2->data[ i ] = (c >= 'a' && c <= 'z') ? c - 32 : c;
	}
#else
	crunch_string s2 = (crunch_string)crunch_alloc(s->len * 4 + 1);
	int n = s->len;
	char *p = s->data, *p2 = s2->data;
	while(n--) {
		uint32_t c;
		p = crunch_utf_decode(p, &c);
		c = crunch_utf_char_upcase(c);
		p2 = crunch_utf_encode(c, p2);
	}
	*(p2++) = '\0';
	int tlen = p2 - s2->data + 1;
	if(tlen < s->size) {
		s2->data = realloc(s2->data, tlen);
		s2->size = tlen;
	}
#endif
	return s2;
}

static crunch_string crunch_string_downcase(crunch_string s) {
#ifdef CRUNCH_NO_UTF
	crunch_string s2 = (crunch_string)crunch_alloc(s->size);
	for(int i = 0; i < s->size; ++i) {
		uint32_t c = s->data[ i ];
		s2->data[ i ] = (c >= 'A' && c <= 'Z') ? c + 32 : c;
	}
#else
	crunch_string s2 = (crunch_string)crunch_alloc(s->len * 4 + 1);
	int n = s->len;
	char *p = s->data, *p2 = s2->data;
	while(n--) {
		uint32_t c;
		p = crunch_utf_decode(p, &c);
		c = crunch_utf_char_downcase(c);
		p2 = crunch_utf_encode(c, p2);
	}
	*(p2++) = '\0';
	int tlen = p2 - s2->data + 1;
	if(tlen < s->size) {
		s2->data = realloc(s2->data, tlen);
		s2->size = tlen;
	}
#endif
	return s2;
}

#ifdef CRUNCH_NO_UTF
# define crunch_string_foldcase		crunch_string_downcase
#else
CRUNCH_EXTERN  crunch_string crunch_utf_string_foldcase(crunch_string s);
# define crunch_string_foldcase		crunch_utf_string_foldcase
#endif

/* (chicken bitwise) */

static inline long crunch_arithmetic_shift(long n, long x) {
	return x > 0 ? n << x : n >> -x;
}

static long crunch_integer_length(long x) {
	/*
	 * From Hacker's Delight by Henry S. Warren
	 * based on a modified nlz() from section 5-3 (fig. 5-7)
	 */
	unsigned long y;
  	long n = 0;
#ifdef CRUCH_64
  	y = x >> 32; if (y != 0) { n += 32; x = y; }
#endif
  	y = x >> 16; if (y != 0) { n += 16; x = y; }
  	y = x >>  8; if (y != 0) { n +=  8; x = y; }
  	y = x >>  4; if (y != 0) { n +=  4; x = y; }
  	y = x >>  2; if (y != 0) { n +=  2; x = y; }
  	y = x >>  1; if (y != 0) return n + 2;
  	return (n + x) < 0 ? ~x : x;
}


/* (chicken memory) */

static inline void *crunch_allocate(int n) { return malloc(n); }
static inline void crunch_free(void *p) { free(p); }
static inline long crunch_pointer_to_address(void *p) { return (long)p; }
static inline void *crunch_address_to_pointer(long a) { return (void *)a; }

static inline long crunch_pointer_u8_ref(void *p) { return *((uint8_t *)p); }
static inline long crunch_pointer_s8_ref(void *p) { return *((int8_t *)p); }
static inline long crunch_pointer_u16_ref(void *p) { return *((uint16_t *)p); }
static inline long crunch_pointer_s16_ref(void *p) { return *((int16_t *)p); }
static inline long crunch_pointer_u32_ref(void *p) { return *((uint32_t *)p); }
static inline long crunch_pointer_s32_ref(void *p) { return *((int32_t *)p); }
static inline long crunch_pointer_u64_ref(void *p) { return *((uint64_t *)p); }
static inline long crunch_pointer_s64_ref(void *p) { return *((int64_t *)p); }
static inline double crunch_pointer_f32_ref(void *p) { return *((float *)p); }
static inline double crunch_pointer_f64_ref(void *p) { return *((double *)p); }
static inline double complex crunch_pointer_c64_ref(void *p) { return *((float complex *)p); }
static inline double complex crunch_pointer_c128_ref(void *p) { return *((double complex *)p); }

static inline void crunch_pointer_u8_set(void *p, long n) { *((uint8_t *)p) = n; }
static inline void crunch_pointer_s8_set(void *p, long n) { *((int8_t *)p) = n; }
static inline void crunch_pointer_u16_set(void *p, long n) { *((uint16_t *)p) = n; }
static inline void crunch_pointer_s16_set(void *p, long n) { *((int16_t *)p) = n; }
static inline void crunch_pointer_u32_set(void *p, long n) { *((uint32_t *)p) = n; }
static inline void crunch_pointer_s32_set(void *p, long n) { *((int32_t *)p) = n; }
static inline void crunch_pointer_u64_set(void *p, long n) { *((uint64_t *)p) = n; }
static inline void crunch_pointer_s64_set(void *p, long n) { *((int64_t *)p) = n; }
static inline void crunch_pointer_f32_set(void *p, float n) { *((float *)p) = n; }
static inline void crunch_pointer_f64_set(void *p, float n) { *((double *)p) = n; }
static inline void crunch_pointer_c64_set(void *p, double complex n) { *((float complex *)p) = n; }
static inline void crunch_pointer_c128_set(void *p, double complex n) { *((double complex *)p) = n; }


/* (scheme process-context) */

static void crunch_exit(int n, ...) {
	va_list va;
	va_start(va, n);
	int r = 0;
	if(n > 0) r = va_arg(va, int);
	crunch_cleanup();
	exit(r);
}

static void crunch_emergency_exit(int n, ...) {
	va_list va;
	va_start(va, n);
	int r = 0;
	if(n > 0) r = va_arg(va, int);
	exit(r);
}

/* (crunch memory) */

#define crunch_pointer_ref(p)	(*(p))
#define crunch_pointer_set(p, x)	(*(p) = (x))

#define crunch_pointer_index_ref(p, i)		((p)[ i ])
#define crunch_pointer_index_set(p, i, x)	((p)[ i ] = (x))

static inline uint32_t crunch_pointer_char_ref(char *p) {
#ifdef CRUNCH_NO_UTF
	return *p;
#else
	uint32_t c;
	crunch_utf_decode(p, &c);
	return c;
#endif
}

static inline void crunch_pointer_char_set(char *p, uint32_t c) {
#ifdef CRUNCH_NO_UTF
	*p = c;
#else
	crunch_utf_encode(c, p);
#endif
}


/* (crunch process-context) */

static crunch_vector crunch_command_line(void) {
	crunch_vector cl = (crunch_vector)crunch_alloc(crunch_process_context.argc *
		sizeof(crunch_string));
	cl->flags |= CRUNCH_NESTEDRC;
	for(int i = 0; i < crunch_process_context.argc; ++i) {
		char *arg = crunch_process_context.argv[ i ];
		crunch_string s = (crunch_string)crunch_alloc(strlen(arg) + 1);
		((crunch_string *)cl->data)[ i ] = s;
#ifdef CRUNCH_NO_UTF
		s->len = s->size - 1;
#else
		s->len = crunch_utf_count(arg, strlen(arg));
#endif
		memcpy(s->data, arg, strlen(arg));
		s->rc = 1;
	}
	cl->len = crunch_process_context.argc;
	return cl;
}

static crunch_string crunch_get_environment_variable(int n, crunch_string var, ...) {
	va_list va;
	va_start(va, var);
	crunch_string def = (crunch_string)&crunch_null_string;
	if(n > 1) def = va_arg(va, crunch_string);
	char *e = getenv(var->data);
	if(e == NULL) return def;
	def = (crunch_string)crunch_alloc(strlen(e) + 1);
#ifdef CRUNCH_NO_UTF
	def->len = def->size - 1;
#else
	def->len = crunch_utf_count(e, strlen(e));
#endif
	memcpy(def->data, e, strlen(e));
	return def;
}


/* (chicken fixnum) */

static inline long crunch_fxrem(long x, long y)  {
	return x - ((x / y) * y);
}

/*
 * From Hacker's Delight by Henry S. Warren
 * based on a modified nlz() from section 5-3 (fig. 5-7)
 */
static inline long crunch_fxlen(long x) {
  unsigned long y;
  long n = 0;

#ifdef CRUNCH_64
  y = x >> 32; if (y != 0) { n += 32; x = y; }
#endif
  y = x >> 16; if (y != 0) { n += 16; x = y; }
  y = x >>  8; if (y != 0) { n +=  8; x = y; }
  y = x >>  4; if (y != 0) { n +=  4; x = y; }
  y = x >>  2; if (y != 0) { n +=  2; x = y; }
  y = x >>  1; if (y != 0) return n + 2;
  return n + x;
}


/* embedded mode */

#ifdef CRUNCH_EMBEDDED
#ifndef CRUNCH_NO_UTF
#include "crunch-utf.c"
#endif

static crunch_string crunch_scheme_string(C_word str) {
	C_word bv = C_block_item(str, 0);
	int len = C_unfix(C_block_item(str, 1));
	int size = C_header_size(bv);
	crunch_string s = (crunch_string)malloc(sizeof(crunch_string_block));
	s->len = len;
	s->flags = CRUNCH_BLOCK;
	s->data = C_data_pointer(bv);
	return s;
}

static crunch_u8vector crunch_scheme_u8vector(C_word vec) {
	int size = C_header_size(vec);
	crunch_u8vector v = (crunch_u8vector)malloc(sizeof(crunch_vector_block));
	v->len = v->size = C_header_size(vec);
	v->flags = CRUNCH_BLOCK;
	v->data = C_data_pointer(vec);
	return v;
}

#define crunch_scheme_make_numvector(t, et) \
static crunch_## t ## vector crunch_scheme_ ## t ## vector(C_word vec) { \
	int size = C_header_size(vec); \
	crunch_ ## t ## vector v = (crunch_ ## t ## vector)malloc(sizeof(crunch_  ## t ## vector)); \
	v->size = C_header_size(vec); \
	v->len = v->size / sizeof(et); \
	v->flags = CRUNCH_BLOCK; \
	v->data = C_data_pointer(vec); \
	return v; \
}

crunch_scheme_make_numvector(s8, int8_t)
crunch_scheme_make_numvector(u16, uint16_t)
crunch_scheme_make_numvector(s16, int16_t)
crunch_scheme_make_numvector(u32, uint32_t)
crunch_scheme_make_numvector(s32, int32_t)
crunch_scheme_make_numvector(u64, uint64_t)
crunch_scheme_make_numvector(s64, int64_t)
crunch_scheme_make_numvector(f32, float)
crunch_scheme_make_numvector(f64, double)
crunch_scheme_make_numvector(c64, float complex)
crunch_scheme_make_numvector(c128, double complex)

#ifdef C_SIXTY_FOUR
#define crunch_scheme_intvector	crunch_scheme_s64vector
#else
#define crunch_scheme_intvector	crunch_scheme_s32vector
#endif

static C_word C_crunch_unref(C_word mptr) {
	crunch_unref_internal((crunch_vector_block *)C_c_pointer_nn(mptr));
	return C_SCHEME_UNDEFINED;
}

static C_word C_crunch_numvector_result(C_word mptr) {
	crunch_vector_block *p = (crunch_vector_block *)C_c_pointer_nn(mptr);
	if((p->flags & CRUNCH_BLOCK) != 0)
		return (C_word)(p - sizeof(C_header));
	C_word b = C_scratch_alloc(C_SIZEOF_BYTEVECTOR(p->size));
	C_block_header_init(b, C_make_header(C_BYTEVECTOR_TYPE, p->size));
	memcpy(C_data_pointer(b), p->data, p->size);
	crunch_unref_internal(p);
	return b;
}

#endif
#endif
